from dataclasses import dataclass
from enum import Enum
from typing import List, Union


@dataclass
class AlignedStep:
    source_idx: int
    target_idx: int


@dataclass
class RangeAlignment:
    source: str
    target: str
    aligned_steps: List[AlignedStep]


@dataclass
class ValueAlignment:
    source: str
    target: str


Alignment = Union[RangeAlignment, ValueAlignment]


class AlignmentType(Enum):
    range = "range"
    value = "value"



from dataclasses import dataclass, asdict
from enum import Enum
from typing import Dict, Union

import ujson, traceback

# noinspection PyUnresolvedReferences
from drepr import drepr_engine
from drepr.version import __engine_version__
from drepr.models import DRepr, DEFAULT_RESOURCE_ID
from drepr.patches import xml_patch, jp_propname_patch

assert drepr_engine.__version__ == __engine_version__, f"You are using a different version of D-REPR" \
                                                          f": {drepr_engine.__version__}. The correct one is" \
                                                          f": {__engine_version__}"


def execute(ds_model: DRepr,
            resources: Union[str, Dict[str, str]],
            output: "Output",
            debug: bool = False):
    ptr = None

    if type(resources) is str:
        resources = {DEFAULT_RESOURCE_ID: resources}

    # ds_model = xml_patch.patch(ds_model, resources)
    # ds_model = jp_propname_patch.patch(ds_model, resources)

    if isinstance(output, FileOutput):
        engine_output = {
            "file": {
                "fpath": str(output.fpath),
                "format": output.format.value
            }
        }
    elif isinstance(output, StringOutput):
        engine_output = {
            "string": {
                "format": output.format.value
            }
        }
    else:
        raise NotImplementedError()

    try:
        ds_model = ds_model.to_engine_format()
        if debug:
            print(">>> the engine is going to execute the below drepr model")
            print(ujson.dumps(ds_model.model, indent=4))

        ptr = drepr_engine.create_executor(ujson.dumps({
            "resources": [
                {"file": resources[rid]}
                for rid in sorted(resources.keys(), key=lambda k: ds_model.resource_idmap[k])
            ],
            "output": engine_output,
            "description": ds_model.model
        }))
        if debug:
            print(f"""
******************************************
>>> Execution plan:
{drepr_engine.get_exec_plan(ptr)}
******************************************
""")

        results = drepr_engine.run_executor(ptr, ujson.dumps(engine_output))
        return results
    finally:
        if ptr is not None:
            drepr_engine.destroy_executor(ptr)


class OutputFormat(Enum):
    TTL = "ttl"
    GraphJSON = "graph_json"


@dataclass
class FileOutput:
    fpath: str
    format: OutputFormat


@dataclass
class StringOutput:
    format: OutputFormat


Output = Union[FileOutput, StringOutput]



from drepr.utils.validator import *
from .align_parser import AlignParser
from .attr_parser import AttrParser
from .preprocessing_parser import PreprocessingParser
from .resource_parser import ResourceParser
from .sm_parser import SMParser
from ..sm import SemanticModel


class ReprV1Parser:
    """
    The DREPR language version 1 has the following schema:

    ```
    version: '1'
    resources: <resources>
    [preprocessing]: <preprocessing> (default is empty list)
    attributes: <attributes>
    [alignments]: <alignments> (default is empty list)
    semantic_model: <semantic_model>
    ```
    """

    TOP_KEYWORDS = {"version", "resources", "preprocessing", "attributes", "alignments", "semantic_model"}
    DEFAULT_RESOURCE_ID = "default"

    @classmethod
    def parse(cls, raw: dict):
        from ..drepr import DRepr

        Validator.must_be_subset(cls.TOP_KEYWORDS, raw.keys(),
                                 setname="Keys of D-REPR configuration",
                                 error_msg="Parsing D-REPR configuration")

        for prop in ['version', 'resources', 'attributes']:
            Validator.must_have(raw, prop, error_msg="Parsing D-REPR configuration")

        Validator.must_equal(raw['version'], '1', "Parsing D-REPR configuration version")
        resources = ResourceParser.parse(raw['resources'])

        if len(resources) == 1:
            default_resource_id = resources[0].id
        else:
            default_resource_id = ResourceParser.DEFAULT_RESOURCE_ID

        preprocessing = PreprocessingParser.parse(default_resource_id, raw.get('preprocessing', []))
        attrs = AttrParser.parse(default_resource_id, raw['attributes'])
        aligns = AlignParser.parse(raw.get('alignments', []))

        if 'semantic_model' in raw:
            sm = SMParser.parse(raw['semantic_model'])
            sm.prefixes.update(SemanticModel.get_default_prefixes())
        else:
            sm = None

        return DRepr(resources, preprocessing, attrs, aligns, sm)



import os
from pathlib import Path

import pytest


def get_examples_dir():
    testdir = Path(os.path.dirname(os.path.abspath(__file__)))
    return testdir.parent / "examples"


@pytest.fixture(params=[
    item for item in get_examples_dir().iterdir()
    if item.is_dir() and not (item / ".ignore").exists()
], ids=lambda item: item.stem)
def example_dir(request) -> Path:
    return request.param



from typing import List

from drepr.utils.validator import Validator, InputError

from ..resource import *


class ResourceParser:
    """
    `resources` has two possible schemas

    1. Shorthand when you have only one resource (`resource_id` is `default`): `resources: <resource_type>`
    2. When you have multiple resources:
        ```
        resources:
            <resource_id>: <resource_conf>
            # .. other resources ..
        ```

        The `<resource_conf>` can either be:
        a. `<resource_type>`, when other resource properties are all options
        b. a dictionary as follows, when some properties require to define explicitly
            ```
                <type>: <resource_type>
                # .. other attributes ..
            ```
    """
    DEFAULT_RESOURCE_ID = "default"
    RESOURCE_TYPES = {rtype.value for rtype in ResourceType}

    @classmethod
    def parse(cls, resources: Union[str, dict]) -> List[Resource]:
        if isinstance(resources, str):
            return cls._parse_schema1(resources)

        if isinstance(resources, dict):
            return cls._parse_schema2(resources)

        raise InputError(
            f"Invalid type for `resources`, expect either a string or a dictionary. Get {type(resources)} instead")

    @classmethod
    def _parse_schema1(cls, resource_type: str) -> List[Resource]:
        Validator.must_in(resource_type, cls.RESOURCE_TYPES, error_msg="Invalid resource type for `resources`")
        resource_type = ResourceType(resource_type)
        if resource_type == ResourceType.CSV:
            resource_prop = CSVProp()
        else:
            resource_prop = None

        return [Resource(cls.DEFAULT_RESOURCE_ID, resource_type, resource_prop)]

    @classmethod
    def _parse_schema2(cls, resources: dict) -> List[Resource]:
        result = []
        for resource_id, conf in resources.items():
            trace = f"Parsing resource {resource_id}."

            if isinstance(conf, str):
                resource = cls._parse_schema1(conf)[0]
                resource.id = resource_id
            elif isinstance(conf, dict):
                Validator.must_have(conf, "type", trace)
                Validator.must_in(conf['type'], cls.RESOURCE_TYPES, f"{trace}\n\tParsing resource type")
                resource_type = ResourceType(conf['type'])

                if resource_type == ResourceType.CSV:
                    if 'delimiter' in conf:
                        if len(conf['delimiter']) != 1:
                            raise InputError(f"{trace}.\nERROR: Expect one character delimiter "
                                             f"for CSV resource. Get `{conf['delimiter']}` instead")

                        resource_prop = CSVProp(conf['delimiter'])
                    else:
                        resource_prop = CSVProp()
                else:
                    resource_prop = None

                resource = Resource(resource_id, resource_type, resource_prop)
            else:
                raise InputError(f"{trace}.\nERROR: The configuration of a resource can either be string "
                                 f"or dictionary. Get {type(conf)} instead")

            result.append(resource)
        return result



from typing import Dict, Union, Iterable

from drepr import DRepr
from drepr.graph import Node, Edge
from drepr.models.align import RangeAlignment
from drepr.models.preprocessing import PreprocessingType, PMap
from drepr.models.resource import ResourceType


class NDArray:

    def __init__(self, ds_model: DRepr, resource: str):
        self.ds_model = ds_model
        self.resource = resource

    def iter_nodes(self) -> Iterable[Node]:
        pass

    def iter_nodes_by_class(self, cls: str) -> Iterable[Node]:
        pass

    def iter_edges(self) -> Iterable[Edge]:
        pass

    @staticmethod
    def from_drepr(ds_model: DRepr, resources: Union[str, Dict[str, str]]):
        # check if we can create a NDArray representation of the dataset
        # the convention that we support is CF convention
        if not NDArray.is_cf_convention(ds_model):
            return False

        if isinstance(resources, dict):
            resource = next(iter(resources.values()))
        else:
            resource = resources

        return NDArray(ds_model, resource)

    @staticmethod
    def is_cf_convention(ds_model: DRepr) -> bool:
        # have only one resource, which is netcdf
        if len(ds_model.resources) > 1 and ds_model.resources[0].type != ResourceType.NetCDF:
            return False

        # only have map preprocessing, which mutate the current data
        for prepro in ds_model.preprocessing:
            if not isinstance(prepro.value, PMap) \
                    or prepro.value.output is not None \
                    or prepro.value.change_structure:
                return False

        # all alignments are dimension alignments
        for align in ds_model.aligns:
            if not isinstance(align, RangeAlignment):
                return False



from collections import defaultdict
from typing import List, Dict, Tuple, Callable, Any, Optional

from drepr.models import DRepr


def patch(repr: DRepr, resources: Dict[str, str]) -> DRepr:
    """
    This patch will fix json path that select property names, before we officially support
    selecting property names

    How does it work: it creates a preprocessing function that convert an object to list of key-value pairs
    """
    need_patch = False
    for pfunc in repr.get_preprocess():
        for slice in pfunc['input']['slices']:
            if slice == '*~':
                raise Exception(
                    "Does not support preprocessing that is applied on properties of an object yet!"
                )

    for attr_name, attr in repr.get_attributes().items():
        if sum(int(slice == '*~') for slice in attr['location']['slices']) > 1:
            raise Exception(f"Invalid path for attribute {attr_name}")

        for slice in attr['location']['slices']:
            if slice == '*~':
                need_patch = True
                break

    if need_patch:
        repr = repr.clone()

        property_paths = []
        for attr_name, attr in repr.get_attributes().items():
            slices = attr['location']['slices']
            for i, slice in enumerate(slices):
                if slice == '*~':
                    property_paths.append((attr_name, i, slices[:i]))
                    break

        # now we need to make sure for every property paths, all paths that is overlap to the property paths,
        # need to be exactly equal and no property paths that is subpath of another property path
        for i in range(len(property_paths)):
            for j in range(i + 1, len(property_paths)):
                if is_path_overlap(property_paths[i][-1], property_paths[j][-1]):
                    raise Exception(
                        f"Does not handle nested selecting property names yet. Found in either {property_paths[i][0]} or {property_paths[i][1]}"
                    )

        overlapped_paths = defaultdict(lambda: [])
        for aid, attr in repr.get_attributes().items():
            slices = attr['location']['slices']
            for i, (paid, _, path) in enumerate(property_paths):
                if len(slices) > len(path) and is_path_overlap(path, slices):
                    assert slices[:len(path
                                       )] == path, f"Slices {slices} does not fully contain {path}"
                    overlapped_paths[i].append(aid)

        # add preprocessing functions
        attrs = repr.get_attributes()
        for paid, _, path in property_paths:
            # need to convert dict to key-value pairs
            repr.get_preprocess().append({
                'type': 'rmap-dict2items',
                'input': {
                    "resource_id": attrs[paid]['location']['resource_id'],
                    "slices": path,
                }
            })

        # now every property path is distinct, and we have a map of property path => attributes
        # that require to re-map their positions
        for pi, aids in overlapped_paths.items():
            _, _, path = property_paths[pi]
            for aid in aids:
                # remap attribute slices
                attr = attrs[aid]
                slices = attr['location']['slices']

                if slices[len(path)] == '*~':
                    # this select property names
                    slices = slices[:len(path)] + ['..', 0] + slices[len(path) + 1:]
                else:
                    assert slices[len(
                        path
                    )] == '..', 'Does not support select both all property names and just a subset of its values'
                    slices = slices[:len(path)] + ['..', 1] + slices[len(path) + 1:]
                attr['location']['slices'] = slices

                # remap alignments
                for alignment in repr.get_alignments():
                    if alignment['type'] == 'dimension':
                        if alignment['source'] == aid:
                            for adim in alignment['aligned_dims']:
                                if adim['source'] > len(path):
                                    adim['source'] += 1
                        elif alignment['target'] == aid:
                            for adim in alignment['aligned_dims']:
                                if adim['target'] > len(path):
                                    adim['target'] += 1

    return repr


def is_path_overlap(path_0: List[str], path_1: List[str]):
    def is_slice_overlap(s0, s1):
        if s0['type'] == 'range':
            if any(str(s0[key]).startswith("${") for key in ('start', 'end', 'step')):
                raise Exception("Haven't support compare expr overlapped yet")

            if s1['type'] == 'range':
                if any(str(s1[key]).startswith("${") for key in ('start', 'end', 'step')):
                    raise Exception("Haven't support compare expr overlapped yet")

                if s1['start'] < (s0['end'] or float('inf')) and (s1['end']
                                                                  or float('inf')) > s0['start']:
                    return True
                return False
            else:
                if isinstance(s1['idx'], str):
                    return False
                elif s0['start'] <= s1['idx'] < (s0['end'] or float('inf')):
                    return True
                return False
        else:
            if s1['type'] == 'range':
                if any(str(s1[key]).startswith("${") for key in ('start', 'end', 'step')):
                    raise Exception("Haven't support compare expr overlapped yet")

                if isinstance(s0['idx'], str):
                    return False
                elif s1['start'] <= s0['idx'] < (s1['end'] or float('inf')):
                    return True
                return False
            else:
                return s1['idx'] == s0['idx']

    slices_0 = [Repr.parse_slice(slice) for slice in path_0]
    slices_1 = [Repr.parse_slice(slice) for slice in path_1]

    if len(slices_0) > len(slices_1):
        slices_0, slices_1 = slices_1, slices_0

    for i in range(len(slices_0)):
        s0 = slices_0[i]
        s1 = slices_1[i]

        if not is_slice_overlap(s0, s1):
            return False

    return True



import abc
import re
from collections import OrderedDict
from typing import Any, Set, Iterable, Dict, Union

import ujson


class InputError(Exception):
    pass


class Validator:
    @staticmethod
    def must_be_dict(val: Any, error_msg: str):
        if not isinstance(val, (dict, OrderedDict)):
            raise InputError(f"{error_msg}\nERROR: Expect a dictionary. Get: {type(val)}")

    @staticmethod
    def must_be_list(val: Any, error_msg: str):
        if not isinstance(val, list):
            raise InputError(f"{error_msg}\nERROR: Expect a list. Get: {type(val)}")

    @staticmethod
    def must_be_str(val: Any, error_msg: str):
        if not isinstance(val, str):
            raise InputError(f"{error_msg}\nERROR: Expect a str. Get: {type(val)}")

    @staticmethod
    def must_be_int(val: Any, error_msg: str):
        if not isinstance(val, int):
            raise InputError(f"{error_msg}\nERROR: Expect a int. Get: {type(val)}")

    @staticmethod
    def must_be_bool(val: Any, error_msg: str):
        if not isinstance(val, bool):
            raise InputError(f"{error_msg}\nERROR: Expect a bool. Get: {type(val)}")

    @staticmethod
    def must_be_subset(parent: Set[Any], subset: Iterable[Any], setname: str, error_msg: str):
        if not parent.issuperset(subset):
            raise InputError(f"{error_msg}\nERROR: {setname.capitalize()} must be a subset of {parent}. Get: {subset}")

    @staticmethod
    def must_in(val: Any, choices: Set[str], error_msg: str):
        if val not in choices:
            raise InputError(
                f"{error_msg}\nERROR: Get `{val}` while possible values are {choices}"
            )

    @staticmethod
    def must_have(odict: dict, attr: str, error_msg: str):
        if attr not in odict:
            raise InputError(
                f"{error_msg}\nERROR: The attribute `{attr}` is missing in the object: `{ujson.dumps(odict, indent=4)}`")

    @staticmethod
    def must_equal(val: Any, expected_val: Any, error_msg: str):
        if val != expected_val:
            raise InputError(f"{error_msg}\nERROR: The value should be: {expected_val}, get: {val} instead")


class SchemaValidator(abc.ABC):

    def __init__(self, is_optional: bool):
        self.is_optional = is_optional

    @abc.abstractmethod
    def validate(self, value):
        raise NotImplementedError()

    @abc.abstractmethod
    def raise_error(self, value):
        raise NotImplementedError()

    @abc.abstractmethod
    def to_string(self):
        raise NotImplementedError()


class DictValidator(SchemaValidator):
    REG_PT = re.compile(r"^int|str|float|any$")
    REG_OPTIONAL = re.compile(r"^optional\((.+)\)$")
    REG_LIST = re.compile(r"^list\((.+)\)$")

    def __init__(self, cls: str, is_optional: bool, **kwargs):
        super().__init__(is_optional)
        self.cls = cls
        self.attrs: Dict[str, SchemaValidator] = {}
        for kw, arg in kwargs.items():
            if isinstance(arg, str):
                m = self.REG_OPTIONAL.match(arg)
                is_optional = False
                if m is not None:
                    is_optional = True
                    arg = m.group(1)

                m = self.REG_LIST.match(arg)
                is_list = False
                if m is not None:
                    is_list = True
                    arg = m.group(1)

                if is_list:
                    m = self.REG_OPTIONAL.match(arg)
                    is_elem_optional = False
                    if m is not None:
                        is_elem_optional = True
                        arg = m.group(1)

                    if arg == "any":
                        self.attrs[kw] = ListValidator(AnyValidator(False), is_optional)
                    else:
                        self.attrs[kw] = ListValidator(
                            PrimitiveValidator(arg, is_elem_optional), is_optional)
                else:
                    if arg == "any":
                        self.attrs[kw] = AnyValidator(False)
                    else:
                        self.attrs[kw] = PrimitiveValidator(arg, is_optional)
            else:
                self.attrs[kw] = arg

        self.attr_names = set(self.attrs.keys())

    def validate(self, odict):
        if self.is_optional and odict is None:
            return

        if not isinstance(odict, (dict, OrderedDict)):
            self.raise_error(odict)

        if not self.attr_names.issuperset(odict.keys()):
            self.raise_error(odict)

        for name, attr in self.attrs.items():
            if name not in odict:
                if not attr.is_optional:
                    self.raise_error(odict)
            else:
                try:
                    attr.validate(odict[name])
                except InputError:
                    self.raise_error(odict)

    def raise_error(self, odict):
        raise InputError(
            f"The schema of object: {odict} does not match with the desired schema: {self.to_string()}"
        )

    def to_string(self):
        return ujson.dumps({k: v.to_string() for k, v in self.attrs.items()}, indent=4)


class AnyValidator(SchemaValidator):

    def validate(self, value):
        pass

    def raise_error(self, value):
        pass

    def to_string(self):
        return "any"


class PrimitiveValidator(SchemaValidator):
    def __init__(self, type_name: str, is_optional: bool):
        super().__init__(is_optional)
        self.type_name = type_name

        if type_name == "str":
            self.type_value = str
        elif type_name == "int":
            self.type_value = int
        elif type_name == "float":
            self.type_value = float
        else:
            raise Exception("Unreachable!")

    def validate(self, value):
        if self.is_optional and value is None:
            return

        if not isinstance(value, self.type_value):
            self.raise_error(value)

    def raise_error(self, odict):
        raise InputError(
            f"The schema of object: {odict} does not match with the desired schema: {self.to_string()}"
        )

    def to_string(self):
        if self.is_optional:
            return f"optional({self.type_name})"
        return self.type_name


class ListValidator(SchemaValidator):
    def __init__(self, element_type: SchemaValidator, is_optional: bool):
        super().__init__(is_optional)
        self.element_type = element_type

    def validate(self, value):
        if self.is_optional and value is None:
            return

        if not isinstance(value, list):
            self.raise_error(value)

        try:
            for v in value:
                self.element_type.validate(v)
        except InputError:
            self.raise_error(value)

    def raise_error(self, odict):
        raise InputError(
            f"The schema of object: {odict} does not match with the desired schema: {self.to_string()}"
        )

    def to_string(self):
        if self.is_optional:
            return f"optional(list({self.element_type.to_string()}))"
        return f"list({self.element_type.to_string()})"



from typing import List, Dict, Tuple, Callable, Any, Optional, Union

import ujson
import xmltodict, uuid
from drepr.models import DRepr


def patch(repr: DRepr, resources: Dict[str, str]) -> DRepr:
    """
    This patch will turn any XML resources to JSON using xmltodict
    """
    need_patch = False
    for rid, resource in repr.get_resources().items():
        if resource['type'] == 'xml':
            need_patch = True

    if need_patch:
        repr = repr.clone()
        for rid, resource in repr.get_resources().items():
            if resource['type'] == 'xml':
                with open(resources[rid], "r") as f:
                    doc = xmltodict.parse(f.read())

                resources[rid] = f"/tmp/{str(uuid.uuid4())}.json"
                resource['type'] = 'json'
                with open(resources[rid], "w") as f:
                    ujson.dump(doc, f)

    return repr


from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

from .path import Path


@dataclass
class PMap:
    resource_id: str
    path: Path
    code: str
    output: Optional[str] = None
    change_structure: Optional[bool] = None


@dataclass
class PFilter:
    resource_id: str
    path: Path
    code: str
    output: Optional[str] = None


class RMapFunc(Enum):
    Dict2Items = "dict2items"


@dataclass
class RMap:
    resource_id: str
    path: Path
    func_id: RMapFunc
    output: Optional[str] = None


class PreprocessingType(Enum):
    pmap = "pmap"
    pfilter = "pfilter"
    rmap = "rmap"


@dataclass
class Preprocessing:
    type: PreprocessingType
    value: Union[PMap, PFilter, RMap]

    @staticmethod
    def deserialize(raw: dict):
        type = PreprocessingType(raw['type'])
        raw['value']['path'] = Path.deserialize(raw['value']['path'])
        if type == PreprocessingType.pmap:
            value = PMap(**raw['value'])
        elif type == PreprocessingType.pfilter:
            value = PFilter(**raw['value'])
        elif type == PreprocessingType.rmap:
            value = RMap(**raw['value'])
        else:
            raise NotImplementedError()

        return Preprocessing(type, value)




from dataclasses import dataclass
from enum import Enum
from typing import List, Dict, Optional, Union

from .attr import Attr


class DataType(Enum):
    xsd_decimal = "xsd:decimal"
    xsd_anyURI = "xsd:anyURI"
    xsd_gYear = "xsd:gYear"
    xsd_dateTime = "xsd:dateTime"
    xsd_int = "xsd:int"
    xsd_string = "xsd:string"


@dataclass
class ClassNode:
    node_id: str
    label: str


@dataclass
class DataNode:
    node_id: str
    attr_id: str
    data_type: Optional[DataType] = None


@dataclass
class LiteralNode:
    node_id: str
    value: str
    data_type: Optional[DataType] = None


@dataclass
class Edge:
    source_id: str
    target_id: str
    label: str
    is_subject: bool = False


Node = Union[LiteralNode, DataNode, ClassNode]


@dataclass
class SemanticModel:
    nodes: Dict[str, Node]
    edges: List[Edge]
    prefixes: Dict[str, str]

    @staticmethod
    def get_default(attrs: List[Attr]) -> 'SemanticModel':
        """
        Automatically generate a semantic model from a list of attributes.

        WARNING: the engine may not able to map data to this semantic model if the final output should be
        comprised of multiple tables.
        """
        prefixes = {"eg": "https://example.org/"}
        aids = {attr.id for attr in attrs}
        cid = None
        for i in range(len(attrs)):
            cid = f"c{i}"
            if cid not in aids:
                break
        nodes = {cid: ClassNode(cid, "eg:Record")}
        edges = []
        for attr in attrs:
            nodes[attr.id] = DataNode(attr.id, attr.id, None)
            edges.append(Edge(cid, attr.id, f"eg:{attr.id}"))

        return SemanticModel(nodes, edges, prefixes)

    @staticmethod
    def get_default_prefixes() -> Dict[str, str]:
        return {
            'drepr': "https://purl.org/drepr/1.0/",
            'rdf': "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            'rdfs': "http://www.w3.org/2000/01/rdf-schema#",
            'owl': "http://www.w3.org/2002/07/owl#"
        }

    @staticmethod
    def deserialize(raw: dict) -> "SemanticModel":
        nodes = {}
        for nid, n in raw['nodes'].items():
            if n['type'] == 'class_node':
                nodes[nid] = ClassNode(n['node_id'], n['label'])
            elif n['type'] == 'data_node':
                nodes[nid] = DataNode(n['node_id'], n['attr_id'], DataType(n['data_type']) if n['data_type'] is not None else None)
            elif n['type'] == 'literal_node':
                nodes[nid] = LiteralNode(n['node_id'], n['value'], DataType(n['data_type']) if n['data_type'] is not None else None)
            else:
                raise NotImplementedError()
        edges = [Edge(**e) for e in raw['edges']]
        return SemanticModel(nodes, edges, raw['prefixes'])

    def remove_node(self, node_id: str):
        self.nodes.pop(node_id)
        for i in range(len(self.edges) - 1, -1, -1):
            if self.edges[i].source_id == node_id or self.edges[i].target_id == node_id:
                self.edges.pop(i)

    def get_rel_iri(self, abs_iri: str) -> str:
        for prefix, uri in self.prefixes.items():
            if abs_iri.startswith(uri):
                return f"{prefix}:{abs_iri.replace(uri, '')}"
        raise ValueError("Create create relative IRI because there is no suitable prefix")

    def get_abs_iri(self, rel_iri: str) -> str:
        prefix, val = rel_iri.split(":", 1)
        if prefix not in self.prefixes:
            raise ValueError(f"Cannot create absolute IRI because the prefix {prefix} does not exist")
        return f"{self.prefixes[prefix]}{val}"

    def is_rel_iri(self, iri: str) -> bool:
        return iri.find(":") != -1
